
#include "Camera2.hpp"

void Camera::Display() {
  if (output == screen) {scare_mouse();}
  stretch_output = false;
  if ((v_zoom_wid != op_wid) || (v_zoom_ht != op_ht)) {stretch_output = true;}
  if (stretch_output) {
    //void stretch_blit(BITMAP *source, BITMAP *dest, int source_x, source_y, source_width, source_height, int dest_x, dest_y, dest_width, dest_height);
    stretch_blit(view , output , v_zoom_tlx , v_zoom_tly , v_zoom_wid , v_zoom_ht , op_tlx , op_tly , op_wid , op_ht);
  } else {
    //void blit(BITMAP *source, BITMAP *dest, int source_x, int source_y, int dest_x, int dest_y, int width, int height);
    blit(view , output , v_tlx , v_tly , op_tlx , op_tly , op_wid , op_ht);
  }
  if (output == screen) {unscare_mouse();}
}

void Camera::View_Settings_Boundary_Check() {
  // make sure that the current view(v_zoom_etc) is within v_xmin , v_xmax , v_ymin , and v_ymax boundaries
  //   it is okay if v_tlx , v_tly , v_wid , v_ht , v_brx , and v_bry are outside of min/max boundaries
  bool zoom_in = false;
  double new_zoom_factor = -1.0;
  int min_x_zoom_hundredths = -1;
  int min_y_zoom_hundredths = -1;
  // if zoomed view is outside boundaries , try moving it inside the boundaries first
  //   if zoomed view is too large for boundaries , increase zoom until it fits
  if ((v_zoom_tlx < v_xmin) || (v_zoom_brx > v_xmax)) {// check x position
    if (v_zoom_wid <= v_xbound_wid) {
      if (v_zoom_tlx < v_xmin) {// slide right
        v_zoom_tlx = v_xmin;
        v_zoom_brx = v_xmin + v_xbound_wid - 1;
      }
      if (v_zoom_brx > v_xmax) {// slide left
        v_zoom_brx = v_xmax;
        v_zoom_tlx = v_xmax - v_xbound_wid + 1;
      }
    } else {// zoom boundaries too large , need to zoom in so find the minimum zoom factor to make the zoomed view fit the boundaries
      zoom_in = true;
      new_zoom_factor = static_cast<double>(v_wid) / static_cast<double>(v_xbound_wid);
      min_x_zoom_hundredths = static_cast<int>(new_zoom_factor*100.0);
      if (static_cast<double>(static_cast<int>(new_zoom_factor*100.0)) != (new_zoom_factor*100.0)) {min_x_zoom_hundredths += 1;}
    }
  }
  if ((v_zoom_tly < v_ymin) || (v_zoom_bry > v_ymax)) {
    if (v_zoom_ht <= v_ybound_ht) {
      if (v_zoom_tly < v_ymin) {// slide down
        v_zoom_tly = v_ymin;
        v_zoom_bry = v_ymin + v_ybound_ht - 1;
      }
      if (v_zoom_bry > v_ymax) {// slide up
        v_zoom_bry = v_ymax;
        v_zoom_tly = v_ymax - v_ybound_ht + 1;
      }
    } else {
      zoom_in = true;
      new_zoom_factor = static_cast<double>(v_ht) / static_cast<double>(v_ybound_ht);
      min_y_zoom_hundredths = static_cast<int>(new_zoom_factor*100.0);
      if (static_cast<double>(static_cast<int>(new_zoom_factor*100.0)) != (new_zoom_factor*100.0)) {min_y_zoom_hundredths += 1;}
    }
  }
  if (zoom_in) {
    // find largest minimum zoom factor
    if (min_x_zoom_hundredths > min_y_zoom_hundredths) {
      new_zoom_factor = static_cast<double>(min_x_zoom_hundredths);
      Set_Zoom_Hundredths(min_x_zoom_hundredths);
    } else {
      new_zoom_factor = static_cast<double>(min_y_zoom_hundredths);
      Set_Zoom_Hundredths(min_y_zoom_hundredths);
    }
  }
}
void Camera::Set_Zoom_Rate_Hundredths(int zoom_rate_100ths) {
  if (zoom_rate_100ths < 1) {zoom_rate_100ths = 1;}
  zoom_rate_hundredths = zoom_rate_100ths;
}
bool Camera::Alter_Zoom(int zoom_100ths) {
  if (zoom_100ths == 0) {return false;}
  if (zoom_100ths < 0) {
    return Zoom_Out(abs(zoom_100ths));
  } else {
    return Zoom_In(zoom_100ths);
  }
  return false;
}
bool Camera::Zoom_In(int zoom_100ths) {
  if (zoom_100ths < 0) {zoom_100ths = 0;}
  return Set_Zoom_Hundredths(zoom_hundredths + zoom_100ths);
}
bool Camera::Zoom_Out(int zoom_100ths) {
  if (zoom_100ths < 0) {zoom_100ths = 0;}
  return Set_Zoom_Hundredths(zoom_hundredths - zoom_100ths);
}
bool Camera::Zoom_In() {
  return Set_Zoom_Hundredths(zoom_hundredths + zoom_rate_hundredths);
}
bool Camera::Zoom_Out() {
  int temp_zoom_hundredths = zoom_hundredths - zoom_rate_hundredths;
  if (temp_zoom_hundredths < 100) {temp_zoom_hundredths = 100;}
  if (temp_zoom_hundredths != zoom_hundredths) {
    return Set_Zoom_Hundredths(temp_zoom_hundredths);
  }
  return false;
}
bool Camera::Set_Zoom_Hundredths(int zoom_100ths) {
  if (zoom_100ths < 100) {zoom_100ths = 100;}
  if (zoom_100ths == zoom_hundredths) {return false;}// nothing to do since zoom didn't change
  double temp_zoom_factor = static_cast<double>(zoom_100ths) / 100.0;
  int temp_zoom_width = static_cast<int>(static_cast<double>(v_wid) / temp_zoom_factor);
  int temp_zoom_height = static_cast<int>(static_cast<double>(v_ht) / temp_zoom_factor);
  if (temp_zoom_width > v_xbound_wid) {return false;}// can't zoom to this size , since it would be larger than the boundary width so return false
  if (temp_zoom_height > v_ybound_ht) {return false;}// can't zoom to this size , since it would be larger than the boundary height so return false
  zoom_factor = temp_zoom_factor;
  zoom_hundredths = zoom_100ths;
  v_zoom_wid = temp_zoom_width;
  v_zoom_ht = temp_zoom_height;
  
//  if (v_zoom_wid % 2 != 0) {v_zoom_wid -= 1;}// without these lines , when zooming in , the screen will shake because the stretch blit
//  if (v_zoom_ht % 2 != 0) {v_zoom_ht -= 1;}// without these lines , when zooming in , the screen will shake because the stretch blit
  
  Set_Zoom_Center(v_desired_center_x , v_desired_center_y);
  return true;
}
void Camera::Scroll_Zoom_Center(int scroll_x , int scroll_y) {
  int new_desired_center_x = v_desired_center_x + scroll_x;
  int new_desired_center_y = v_desired_center_y + scroll_y;
  if (new_desired_center_x < v_xmin) {new_desired_center_x = v_xmin;}
  if (new_desired_center_y < v_ymin) {new_desired_center_y = v_ymin;}
  if (new_desired_center_x > v_xmax) {new_desired_center_x = v_xmax;}
  if (new_desired_center_y > v_ymax) {new_desired_center_y = v_ymax;}
  Set_Zoom_Center(new_desired_center_x , new_desired_center_y);
}
void Camera::Set_Zoom_Center(int desired_centerx , int desired_centery) {
  if (log_ptr != NULL) {*log_ptr << "Set_Zoom_Center Called , Desired Center = (" << desired_centerx << "," << desired_centery << ")" << endl;}
///###Wrong , zoom dimensions might have changed
///  if ((desired_centerx == v_desired_center_x) && (desired_centery == v_desired_center_y)) {return;}// center didn't change , nothing to do
//  if (desired_centerx < 0) {desired_centerx = 0;}
//  if (desired_centerx >= view->w) {desired_centerx = view->w - 1;}
//  if (desired_centery < 0) {desired_centery = 0;}
//  if (desired_centery >= view->h) {desired_centery = view->h - 1;}
  if (desired_centerx < v_xmin) {desired_centerx = v_xmin;}
  if (desired_centerx > v_xmax) {desired_centerx = v_xmax;}
  if (desired_centery < v_ymin) {desired_centery = v_ymin;}
  if (desired_centery > v_ymax) {desired_centery = v_ymax;}
  if (desired_centerx < v_zoom_tlx) {desired_centerx = v_zoom_tlx;}
  if (desired_centerx > v_zoom_brx) {desired_centerx = v_zoom_brx;}
  if (desired_centery < v_zoom_tly) {desired_centery = v_zoom_tly;}
  if (desired_centery > v_zoom_bry) {desired_centery = v_zoom_bry;}
  v_desired_center_x = desired_centerx;
  v_desired_center_y = desired_centery;
  int temp_zoom_tlx = desired_centerx - v_zoom_wid / 2;
  int temp_zoom_tly = desired_centery - v_zoom_ht / 2;
  int temp_zoom_brx = temp_zoom_tlx + v_zoom_wid - 1;
  int temp_zoom_bry = temp_zoom_tly + v_zoom_ht - 1;
  if (temp_zoom_tlx < v_xmin) {// too far left of boundary so slide right
    temp_zoom_tlx = v_xmin;
    temp_zoom_brx = temp_zoom_tlx + v_zoom_wid - 1;
  }
  if (temp_zoom_brx > v_xmax) {// too far right of boundary so slide left
    temp_zoom_brx = v_xmax;
    temp_zoom_tlx = temp_zoom_brx - v_zoom_wid + 1;
  }
  if (temp_zoom_tly < v_ymin) {// too far above boundary so slide down
    temp_zoom_tly = v_ymin;
    temp_zoom_bry = temp_zoom_tly + v_zoom_ht - 1;
  }
  if (temp_zoom_bry > v_ymax) {// too far below boundary so slide up
    temp_zoom_bry = v_ymax;
    temp_zoom_tly = temp_zoom_bry - v_zoom_ht + 1;
  }
  v_zoom_tlx = temp_zoom_tlx;
  v_zoom_tly = temp_zoom_tly;
  v_zoom_brx = temp_zoom_brx;
  v_zoom_bry = temp_zoom_bry;
  v_center_x = v_zoom_tlx + v_zoom_wid / 2;
  v_center_y = v_zoom_tly + v_zoom_ht / 2;
  v_tlx = v_center_x - v_wid / 2;
  v_brx = v_tlx + v_wid - 1;
  v_tly = v_center_y - v_ht / 2;
  v_bry = v_tly + v_ht - 1;
  if (log_ptr != NULL) {*log_ptr << "After calling Set Zoom Center , This Camera is : " << *this << endl;}
}
bool Camera::Set_Zoom_Rectangle(int leftx , int rightx , int topy , int bottomy) {
  if (leftx < v_xmin) {return false;}
  if (rightx > v_xmax) {return false;}
  if (topy < v_ymin) {return false;}
  if (bottomy > v_ymax) {return false;}
  v_zoom_wid = rightx - leftx + 1;
  v_zoom_ht = bottomy - topy + 1;
  zoom_factor = static_cast<double>(v_wid) / static_cast<double>(v_zoom_wid);
  zoom_hundredths = static_cast<int>(zoom_factor*100.0);
  // (v_wid / zoom_factor = v_zoom_wid)
  Set_Zoom_Center(leftx + v_zoom_wid/2 , topy + v_zoom_ht / 2);
  return true;
}

bool Camera::Set_View_Boundaries(int xmin , int xmax , int ymin , int ymax) {// don't use until the view bitmap has been setup
  if (view == NULL) {return false;}
  if (xmin < 0) {xmin = 0;}
  if (xmax >= view->w) {xmax = view->w - 1;}
  if (ymin < 0) {ymin = 0;}
  if (ymax >= view->h) {ymax = view->h - 1;}
  if (xmin >= xmax) {return false;}
  if (ymin >= ymax) {return false;}
  v_xmin = xmin;v_xmax = xmax;
  v_ymin = ymin;v_ymax = ymax;
  v_xbound_wid = v_xmax - v_xmin + 1;
  v_ybound_ht = v_ymax - v_ymin + 1;
  
  View_Settings_Boundary_Check();
  
  return true;
}
bool Camera::Map_Output_Coord_to_View_Coord(int op_x , int op_y , int* v_x , int* v_y) {
  if ((op_wid <= 1) || (op_ht <= 1)) {return false;}
  double xscale = static_cast<double>(v_zoom_wid - 1) / static_cast<double>(op_wid - 1);
  double yscale = static_cast<double>(v_zoom_ht - 1) / static_cast<double>(op_ht - 1);
  *v_x = v_zoom_tlx + static_cast<int>(static_cast<double>(op_x - op_tlx)*xscale);
  *v_y = v_zoom_tly + static_cast<int>(static_cast<double>(op_y - op_tly)*yscale);
//      *v_x = v_zoom_tlx + ((op_x - op_tlx) / (op_wid - 1))*(v_zoom_wid - 1);
  return true;
}
Camera::Camera(BITMAP* view_bitmap , BITMAP* output_bitmap) {
  Clear_Camera_Settings();
  if (view_bitmap != NULL) {
    view = view_bitmap;
    v_tlx = 0;v_tly = 0;
    v_wid = view->w;v_ht = view->h;
    v_brx = v_wid - 1;v_bry = v_ht - 1;
    v_center_x = v_tlx + v_wid/2;
    v_center_y = v_tly + v_ht/2;
    v_xmin = v_tlx;v_xmax = v_brx;
    v_ymin = v_tly;v_ymax = v_bry;
    v_xbound_wid = v_wid;
    v_ybound_ht = v_ht;
    v_zoom_tlx = v_tlx;v_zoom_tly = v_tly;
    v_zoom_wid = v_wid;v_zoom_ht = v_ht;
    v_zoom_brx = v_brx;v_zoom_bry = v_bry;
    v_desired_center_x = v_center_x;
    v_desired_center_y = v_center_y;
  }
  if (output_bitmap != NULL) {
    output = output_bitmap;
    op_tlx = 0;op_tly = 0;
    op_wid = output->w;
    op_ht = output->h;
    op_zoomout_tlx = op_tlx;
    op_zoomout_tly = op_tly;
    op_zoomout_wid = op_wid;
    op_zoomout_ht = op_ht;
    op_zoomout_brx = op_tlx + op_wid - 1;
    op_zoomout_bry = op_tly + op_ht - 1;
  }
}
void Camera::Clear_Camera_Settings() {
  log_ptr = NULL;
  view = NULL;
  v_tlx = v_tly = v_wid = v_ht = v_brx = v_bry = v_center_x = v_center_y = 0;
  v_xmin = v_xmax = v_ymin = v_ymax = 0;v_xbound_wid = 0;v_ybound_ht = 0;
  zoom_factor = 1.0;
  zoom_hundredths = 100;
  zoom_rate_hundredths = 1;
  v_zoom_tlx = v_zoom_tly = v_zoom_wid = v_zoom_ht = v_zoom_brx = v_zoom_bry = 0;
  v_desired_center_x = v_desired_center_y = 0;
  
  output = NULL;
  op_zoomout_tlx = op_zoomout_tly = op_zoomout_wid = op_zoomout_ht = op_zoomout_brx = op_zoomout_bry = 0;
  op_zoomout_factor = 1.0;
  op_zoomout_hundredths = 100;
  op_zoomout_rate_hundredths = 1;
  
  stretch_output = false;
//      allow_zoom_out_to_push_view = true;
  allow_recenter_to_change_zoom = false;
}

bool Camera::Set_Output_Rectangle(int leftx , int rightx , int topy , int bottomy) {
  if (leftx < 0) {return false;}
  if (rightx > output->w - 1) {return false;}
  if (topy < 0) {return false;}
  if (bottomy > output->h -1) {return false;}
  op_tlx = leftx;
  op_wid = rightx - leftx + 1;
  op_tly = topy;
  op_ht = bottomy - topy + 1;
  return true;
}



/// ##################### old version ##########################
//#include "Camera.hpp"


/*
void Camera::CheckViewBoundaries() {
  if (v_tlx < v_xmin) {v_tlx = v_xmin;}
  if ((v_tlx + v_wid - 1) > v_xmax) {
    v_wid = v_xmax - v_tlx + 1;
  }
  v_center_x = v_tlx + v_wid / 2;
  
  if (v_tly < v_ymin) {v_tly = v_ymin;}
  if ((v_tly + v_ht - 1) > v_ymax) {
    v_ht = v_ymax - v_tly + 1;
  }
  v_center_y = v_tly + v_ht / 2;
  RenewZoomSpecs();
}

void Camera::SetViewCentered(BITMAP* view_bitmap , int view_centerx , int view_centery , int view_width , int view_height , int view_zoom) {
  view = view_bitmap;
  if (view_zoom < 1) {view_zoom = 1;}
  zoom = view_zoom;
  v_center_x = view_centerx;
  v_center_y = view_centery;
  v_wid = view_width;
  v_ht = view_height;
  v_tlx = v_center_x - v_wid / 2;
  v_tly = v_center_y - v_ht / 2;
  RenewZoomSpecs();
}

void Camera::SetView(BITMAP* view_bitmap , int view_topleftx , int view_toplefty , int view_width , int view_height , int view_zoom) {
  view = view_bitmap;
  if (view_zoom < 1) {view_zoom = 1;}
  zoom = view_zoom;
  v_tlx = view_topleftx;
  v_tly = view_toplefty;
  v_wid = view_width;
  v_ht = view_height;
  v_center_x = v_tlx + v_wid / 2;
  v_center_y = v_tly + v_ht / 2;
  RenewZoomSpecs();
//      v_zoom_wid = v_wid / zoom;
//      v_zoom_ht = v_ht / zoom;
//      v_zoom_tlx = v_center_x - v_zoom_wid / 2;
//      v_zoom_tly = v_center_y - v_zoom_ht / 2;
}

void Camera::CenterViewOnPlayer() {
  if (pl != NULL) {
    v_center_x = pl->x + pl->wd / 2;
    v_center_y = pl->y + pl->ht / 2;
    v_tlx = v_center_x - v_wid / 2;
    v_tly = v_center_y - v_ht / 2;
    CheckViewBoundaries();
    
//    v_zoom_tlx = v_center_x - v_zoom_wid / 2;
//    v_zoom_tly = v_center_y - v_zoom_ht / 2;
  }
}

void Camera::CenterView(int cx , int cy) {
  v_center_x = pl->x + pl->wd / 2;
  v_center_y = pl->y + pl->ht / 2;
  v_zoom_tlx = v_center_x - v_zoom_wid / 2;
  v_zoom_tly = v_center_y - v_zoom_ht / 2;
}

void Camera::RenewZoomSpecs() {
  v_zoom_wid = v_wid / zoom;
  v_zoom_ht = v_ht / zoom;
  v_zoom_tlx = v_center_x - v_zoom_wid / 2;
  v_zoom_tly = v_center_y - v_zoom_ht / 2;
  IsOutputStretched();
}

void Camera::SetOutput(BITMAP* output_bitmap , int output_topleftx , int output_toplefty , int output_width , int output_height) {
  output = output_bitmap;
  op_tlx = output_topleftx;
  op_tly = output_toplefty;
  op_wid = output_width;
  op_ht = output_height;
  IsOutputStretched();
}

void Camera::ShowOutput() {
  if (output == screen) {scare_mouse();}
  if (stretch_output) {
//void stretch_blit(BITMAP *source, BITMAP *dest, int source_x, source_y, source_width, source_height, int dest_x, dest_y, dest_width, dest_height);
    stretch_blit(view , output , v_zoom_tlx , v_zoom_tly , v_zoom_wid , v_zoom_ht , op_tlx , op_tly , op_wid , op_ht);
  } else {
//void blit(BITMAP *source, BITMAP *dest, int source_x, int source_y, int dest_x, int dest_y, int width, int height);
    blit(view , output , v_tlx , v_tly , op_tlx , op_tly , op_wid , op_ht);
  }
  if (output == screen) {unscare_mouse();}
}
*/



