
#ifndef Camera_H
#define Camera_H

#include <allegro.h>
#include <iostream>
#include <fstream>

using std::ofstream;
using std::ostream;
using std::flush;
using std::endl;

//#include "Player.hpp"

class Camera {
  /// Only blit can use the same destination and source
  /// Stretch blit source bitmap must be a memory bitmap , can't copy from outside the source , 
  ///   and source and destination must be different
  public :
    ofstream* log_ptr;
    
    BITMAP* view;
    int v_tlx , v_tly;// position boundaries for zoom_factor = 1.0
    int v_wid , v_ht;// position boundaries for zoom_factor = 1.0
    int v_brx , v_bry;// position boundaries for zoom_factor = 1.0
    
    int v_xmin , v_xmax;// position limiters for left/right view boundaries , default to 0 , view->w
    int v_ymin , v_ymax;// position limiters for top/bottom view boundaries , default to 0 , view->h
    int v_xbound_wid;
    int v_ybound_ht;
    
    double zoom_factor;// 1.0 or larger
    int zoom_hundredths;// 100 or larger
    int zoom_rate_hundredths;// 1 or more , this is how many hundredths the zoom will change by when zoom_in() or zoom_out() is called
    int v_zoom_tlx , v_zoom_tly;// what is actually in view
    int v_zoom_wid , v_zoom_ht;// what is actually in view
    int v_zoom_brx , v_zoom_bry;// what is actually in view
    int v_center_x , v_center_y;// actual center position for any zoom_factor
    
    int v_desired_center_x , v_desired_center_y;// so positions near edges/corners of view bitmap can be centered on without going out of bounds
                                                //    however , this does not mean that the desired center will be the actual center of the view
                                                //    The desired center must be within the bitmap boundaries and will also be constrained to 
                                                //      within the user set boundaries
    
    BITMAP* output;
    int op_tlx;
    int op_tly;
    int op_wid;
    int op_ht;
    
    int op_zoomout_tlx;
    int op_zoomout_tly;
    int op_zoomout_wid;
    int op_zoomout_ht;
    int op_zoomout_brx;
    int op_zoomout_bry;
    
    double op_zoomout_factor;
    int op_zoomout_hundredths;
    int op_zoomout_rate_hundredths;
    
    
    bool stretch_output;// if the size of the zoomed view and the size of the output area are not the same , the view is stretched
//    bool allow_zoom_out_to_push_view;// if zoomed in near corner and zooming out would display out of bounds portion of view_bitmap then
                                 //    view gets pushed in towards the center to accomodate the change in zoom level
                                 //    so if this value is false then zooming out in that situation would not be allowed
    bool allow_recenter_to_change_zoom;// if recentering the current zoom window moves part of it out of bounds , then zoom in to compensate
//    Player* pl;
    void Display();
    
    void View_Settings_Boundary_Check();
    
    void Set_Zoom_Rate_Hundredths(int zoom_rate_100ths);
    bool Alter_Zoom(int zoom_100ths);
    bool Zoom_In(int zoom_100ths);
    bool Zoom_Out(int zoom_100ths);
    bool Zoom_In();
    bool Zoom_Out();
    bool Set_Zoom_Hundredths(int zoom_100ths);
    void Scroll_Zoom_Center(int scroll_x , int scroll_y);
    void Set_Zoom_Center(int desired_centerx , int desired_centery);
    bool Set_Zoom_Rectangle(int leftx , int rightx , int topy , int bottomy);
    bool Set_View_Boundaries(int xmin , int xmax , int ymin , int ymax);// don't use until the view bitmap has been setup
    
    bool Set_Output_Rectangle(int leftx , int rightx , int topy , int bottomy);
    
    void SetLogFile(ofstream* log) {
      log_ptr = log;
    }
    
    bool Map_Output_Coord_to_View_Coord(int op_x , int op_y , int* v_x , int* v_y);
    
    Camera(BITMAP* view_bitmap , BITMAP* output_bitmap);
    Camera() {Clear_Camera_Settings();}
    void Clear_Camera_Settings();
    
    friend ostream& operator<< (ostream& os , const Camera& src ) {
      os << "View { Top Left (" << src.v_tlx << "," << src.v_tly << ") Dimensions (" << src.v_wid << "," << src.v_ht << ") Bottom Right (" << src.v_brx << "," << src.v_bry << ") } ";
      os << "View Boundaries { Xmin = " << src.v_xmin << " , Xmax = " << src.v_xmax << " , Ymin = " << src.v_ymin << " , Ymax = " << src.v_ymax << " } ";
      os << "Zoomed View { Top Left (" << src.v_zoom_tlx << "," << src.v_zoom_tly << ") Dimensions (" << src.v_zoom_wid << "," << src.v_zoom_ht << ") Bottom Right (" << src.v_zoom_brx << "," << src.v_zoom_bry << ") Center (" << src.v_center_x << "," << src.v_center_y << ") } ";
      os << "Output { Top Left (" << src.op_tlx << "," << src.op_tly << ") Dimensions (" << src.op_wid << "," << src.op_ht << ") } ";
      os.flush();
      return os;
    }
    
    
    
    
    /// ########### old ################///
/*
    Camera() {
      v_tlx = v_tly = v_zoom_tlx = v_zoom_tly = v_wid = v_ht = v_zoom_wid = v_zoom_ht = v_center_x = v_center_y = 0;
      v_xmin = v_xmax = v_ymin = v_ymax = 0;
      v_desired_center_x = 0;
      v_desired_center_y = 0;
      zoom = 1;
      view = output = NULL;
      stretch_output = false;
    }
    ~Camera() {}
    
//    void SetView(BITMAP* view_bitmap) {view = view_bitmap;}
//    void SetPlayer(Player* p) {pl = p;CenterViewOnPlayer();}
    
    void SetViewBoundaries(int xmin , int xmax , int ymin , int ymax) {
      v_xmin = xmin;
      v_xmax = xmax;
      v_ymin = ymin;
      v_ymax = ymax;
      CheckViewBoundaries();
    }
    
    void CheckViewBoundaries();
    
    void SetViewCentered(BITMAP* view_bitmap , int desired_view_centerx , int desired_view_centery , int view_width , int view_height , int view_zoom);
    void SetViewCentered(BITMAP* view_bitmap , int desired_view_centerx , int desired_view_centery , int view_width , int view_height , int view_zoom) {
      if (view_bitmap == NULL) {return;}
      if (v_xmin < 0) {v_xmin = 0;}
      if (v_xmax >= view_bitmap->w) {v_xmax = view_bitmap->w - 1;}
      if (v_ymin < 0) {v_ymin = 0;}
      if (v_ymax >= view_bitmap->h) {v_ymax = view_bitmap->h - 1;}
      if (v_xmax < v_xmin) {v_xmax = v_xmin;}
      if (v_ymax < v_ymin) {v_ymax = v_ymin;}
      if (v_tlx < v_xmin) {v_tlx = v_xmin;}
      if (v_tly < v_ymin) {v_tly = v_ymin;}
      if ( (v_wid > (v_xmax - v_tlx)) {
      
      
    }
    void SetView(BITMAP* view_bitmap , int view_topleftx , int view_toplefty , int view_width , int view_height , int view_zoom);
    
    void CenterViewOnPlayer();
    void CenterView(int cx , int cy);
    void ZoomIn() {zoom++;RenewZoomSpecs();}
    void ZoomOut() {if (zoom > 1) {zoom--;RenewZoomSpecs();}}
    void SetZoom(int new_zoom) {if (new_zoom > 0) {zoom = new_zoom;RenewZoomSpecs();}}

    void RenewZoomSpecs();
    
    
    
    void SetOutput(BITMAP* output_bitmap) {output = output_bitmap;}
    
    void SetOutput(BITMAP* output_bitmap , int output_topleftx , int output_toplefty , int output_width , int output_height);
    void PositionTopLeftOfOutPut(int opx , int opy) {
      op_tlx = opx;
      op_tly = opy;
    }
    void PositionCenterOfOutput(int opcx , int opcy) {
      op_tlx = opcx - op_wid / 2;
      op_tly = opcy - op_ht / 2;
    }
    
    void IsOutputStretched() {
      stretch_output = false;
      if ( (v_zoom_wid != op_wid) || (v_zoom_ht != op_ht) ) {stretch_output = true;}
    }
    
    void ShowOutput();
*/
};




#endif





