

#include <cmath>
#include <sstream>

#include "KMaze.hpp"
#include "Shuffle.hpp"

using std::stringstream;


/// #####################   KMaze class member functions   ###################################
KMaze::KMaze() {
  width = 0;
  height = 0;
  sg_width = 0;
  sg_height = 0;
  top_left_x = 0;
  top_left_y = 0;
  
  total_vertice_count = 0;
  VerticeArray = NULL;
  
  total_edge_count = 0;
  AllEdgeArray = NULL;
  
  wemm.clear();
  EdgesOnArray = NULL;
}

KMaze::~KMaze() {
  
}


bool KMaze::InitializeGrid(int w , int h , int tlx , int tly , int grid_z_level , int xwid , int yht) {
  if ((w < 1) || (h < 1)) {return false;}
  if ((xwid < 1) || (yht < 1)) {return false;}
  
  srand(time(NULL));
  
  top_left_x = tlx;
  top_left_y = tly;
  
  sg_width = xwid;
  sg_height = yht;
  
  width = w;
  height = h;
  total_vertice_count = (h+1)*(w+1);
  VerticeArray = new Vertice[total_vertice_count];
  if (VerticeArray == NULL) {return false;}
  
  
  int row = 0;
  int col = 0;
  
  Vertice* V = NULL;
  int vxpos = 0;
  int vypos = 0;
  // build a grid of vertices
  for (row = 0 ; row < h + 1; row++) {
    vypos = tly + row*yht;
    for (col = 0 ; col < w + 1; col++) {
      vxpos = tlx + col*xwid;
/*
          vypos = tly + row*yht;
          vypos += (rand() % ((6*yht)/10)) - (3*yht)/10;// +/- up to 30% of the yht
          vxpos = tlx + col*xwid;
          vxpos += (rand() % ((6*xwid)/10)) - (3*xwid)/10;// +/- up to 30% of the xwid
//*/
      V = &(VerticeArray[(row*(w+1)) + col]);
      V->x = vxpos;
      V->y = vypos;
//      V->z = grid_z_level;
      V->z = grid_z_level + (rand() % 51) - 25;// randomly +/- up to 25 from grid_z_level for the z coordinate
    }
  }
  
  total_edge_count = 2*(w*h) + (w + h);
  AllEdgeArray = new Edge[total_edge_count];
  if (AllEdgeArray == NULL) {
    delete [] VerticeArray;
    return false;
  }
  
  int edge_count = 0;
  Edge* KE = NULL;
  
  
  int edge_row = 0;
  int edge_col = 0;
  int v_row = 0;
  int v_col = 0;

  // process edges in "rows"
  // (zero-based) even rows are the left-right edges
  // odd rows are the up-down edges
  for (edge_row = 0 ; edge_row < (2*h + 1) ; edge_row++) {
    switch (edge_row % 2) {
      // left-right edges
      // w columns
      case 0 :
        v_row = edge_row / 2;
        for (edge_col = 0 ; edge_col < w ; edge_col++) {
          v_col = edge_col;
          KE = &AllEdgeArray[edge_count];
          KE->KWeight = 1;
//          KE->KWeight = rand() % 3;
//          KE->KWeight = rand() % (edge_col + 1);
//          KE->KWeight = static_cast<int>(sqrt(static_cast<double>(abs((edge_row-h/2)*(edge_row-h/2)) + abs((edge_col-w/2)*(edge_col-w/2)))));
//          KE->KWeight = static_cast<int>(sqrt(double(h*h + w*w)/4.0) - sqrt(static_cast<double>(abs((edge_row-h/2)*(edge_row-h/2)) + abs((edge_col-w/2)*(edge_col-w/2)))));
//              KE->vpt1 = &(VerticeArray[v_row][v_col]);
          KE->vpt1 = &(VerticeArray[v_row*(w+1) + v_col]);
//              KE->vpt1 = &(VerticeArray[v_row][v_col + 1]);
          KE->vpt2 = &(VerticeArray[v_row*(w+1) + v_col + 1]);
          edge_count++;
        }
        break;
      // top-bottom edges
      // w + 1 columns
      case 1 :
        v_row = edge_row / 2;
        for (edge_col = 0 ; edge_col < w + 1 ; edge_col++) {
          v_col = edge_col;
          KE = &(AllEdgeArray[edge_count]);
          KE->KWeight = 1;
//          KE->KWeight = rand() % 3;
//          KE->KWeight = rand() % (edge_row + 1);
//          KE->KWeight = static_cast<int>(sqrt(static_cast<double>(abs((edge_row-h/2)*(edge_row-h/2)) + abs((edge_col-w/2)*(edge_col-w/2)))));
//          KE->KWeight = static_cast<int>(sqrt(double(h*h + w*w)/4.0) - sqrt(static_cast<double>(abs((edge_row-h/2)*(edge_row-h/2)) + abs((edge_col-w/2)*(edge_col-w/2)))));
//              KE->vpt1 = &(VerticeArray[v_row][v_col]);
          KE->vpt1 = &(VerticeArray[v_row*(w+1) + v_col]);
//              KE->vpt1 = &(VerticeArray[v_row + 1][v_col]);
          KE->vpt2 = &(VerticeArray[(v_row + 1)*(w+1) + v_col]);
          edge_count++;
        }
        break;
    }
  }
  
  // now that all the edges have been declared , give each vertice a link to any edge
  //   that 'owns' it if it doesn't already
  int e = 0;
  for (e = 0 ; e < total_edge_count ; e++) {
    KE = &(AllEdgeArray[e]);
    
    V = KE->vpt1;
    if (!(V->check_for_edge(KE))) {
      V->assign_edge(KE);
    }
    
    V = KE->vpt2;
    if (!(V->check_for_edge(KE))) {
      V->assign_edge(KE);
    }
  }
  
  
  return true;
}


void KMaze::DestroyGrid() {
  if (VerticeArray != NULL) {
    delete [] VerticeArray;
    VerticeArray = NULL;
  }
  if (AllEdgeArray != NULL) {
    delete [] AllEdgeArray;
    AllEdgeArray = NULL;
  }
}

void KMaze::ResetGridEdges () {
  Edge* KE = NULL;
  for (int i = 0 ; i < total_edge_count ; i++) {
    KE = &(AllEdgeArray[i]);
    KE->checked = false;
    KE->on = false;
  }
}


bool KMaze::SetMazeBorderOn () {
  /// Set all outside edges of the maze to 'on = true' and 'checked = true'
  
  Edge* KE = NULL;
  
  for (int x = 1 ; x <= width ; x++) {
    // set top and bottom edges to on and checked
    if (this->GridCoordToEdgePtr(x , 1 , "top" , &KE)) {
      KE->SetEdgeStatus(true);
    } else {
      return false;// failed to get a pointer to the proper edge
    }
    if (this->GridCoordToEdgePtr(x , height , "bottom" , &KE)) {
      KE->SetEdgeStatus(true);
    } else {
      return false;// failed to get a pointer to the proper edge
    }
  }
  for (int y = 1 ; y <= height ; y++) {
    // set left and right edges to on and checked
    if (GridCoordToEdgePtr(1 , y , "left" , &KE)) {
      KE->SetEdgeStatus(true);
    } else {
      return false;// failed to get a pointer to the proper edge
    }
    if (GridCoordToEdgePtr(width , y , "right" , &KE)) {
      KE->SetEdgeStatus(true);
    } else {
      return false;// failed to get a pointer to the proper edge
    }
  }
  return true;
}

bool KMaze::SetExit(int exit_gx , int exit_gy , string exit_side) {
  /// check to make sure that exits to maze are on the edge
  if ( (exit_gx != 1) && (exit_gx != width) ) {
    if ( (exit_gy != 1) && (exit_gy != height) ) {
      return false;// exit's grid coordinate is in the middle of the grid
    }
  }
  /// check to make sure that the "side" of the the grid coordinate's exit is on the outside edge
  if (exit_gx == 1) {
    if (exit_gy == 1) {
      // grid coordinate 1,1 is top left corner
      if ( (exit_side != "top") && (exit_side != "left") ) {return false;}
    } else if (exit_gy == height) {
      // grid coordinate 1,height is bottom left corner
      if ( (exit_side != "bottom") && (exit_side != "left") ) {return false;}
    } else {
      // grid coordinates 1 , (2 through height - 1) is left side
      if (exit_side != "left") {return false;}
    }
  } else if (exit_gx == width) {
    if (exit_gy == 1) {
      // grid coordinate width,1 is top right corner
      if ( (exit_side != "top") && (exit_side != "right") ) {return false;}
    } else if (exit_gy == height) {
      // grid coordinate width,height is bottom right corner
      if ( (exit_side != "bottom") && (exit_side != "right") ) {return false;}
    } else {
      // grid coordinates width , (2 through height - 1) is left side
      if (exit_side != "right") {return false;}
    }
  }
  
  Edge* KE = NULL;
  
  /// Now set the exit's edge to on = false , and checked = true
  if (GridCoordToEdgePtr(exit_gx , exit_gy , exit_side , &KE)) {
    KE->SetEdgeStatus(false);
  } else {
    return false;// failed to get a pointer to the proper edge
  }
  
  return true;
}

bool KMaze::SetMazeBorderOn (int exit1_gx , int exit1_gy , string side_exit1 , int exit2_gx , int exit2_gy , string side_exit2) {
  
  if ( (exit1_gx < 1) || (exit1_gx > width) ) {return false;}
  if ( (exit1_gy < 1) || (exit1_gy > height) ) {return false;}
  if ( (exit2_gx < 1) || (exit2_gx > width) ) {return false;}
  if ( (exit2_gy < 1) || (exit2_gy > height) ) {return false;}
  
  /// check to make sure that exits to maze are on the edge
  if ( (exit1_gx != 1) && (exit1_gx != width) ) {
    if ( (exit1_gy != 1) && (exit1_gy != height) ) {
      return false;// exit 1's grid coordinate is in the middle of the grid
    }
  }
  if ( (exit2_gx != 1) && (exit2_gx != width) ) {
    if ( (exit2_gy != 1) && (exit2_gy != height) ) {
      return false;// exit 2's grid coordinate is in the middle of the grid
    }
  }
  
  /// check to make sure that the "side" of the the first grid coordinate's exit is on the outside edge
  if (exit1_gx == 1) {
    if (exit1_gy == 1) {
      // grid coordinate 1,1 is top left corner
      if ( (side_exit1 != "top") && (side_exit1 != "left") ) {return false;}
    } else if (exit1_gy == height) {
      // grid coordinate 1,height is bottom left corner
      if ( (side_exit1 != "bottom") && (side_exit1 != "left") ) {return false;}
    } else {
      // grid coordinates 1 , (2 through height - 1) is left side
      if (side_exit1 != "left") {return false;}
    }
  } else if (exit1_gx == width) {
    if (exit1_gy == 1) {
      // grid coordinate width,1 is top right corner
      if ( (side_exit1 != "top") && (side_exit1 != "right") ) {return false;}
    } else if (exit1_gy == height) {
      // grid coordinate width,height is bottom right corner
      if ( (side_exit1 != "bottom") && (side_exit1 != "right") ) {return false;}
    } else {
      // grid coordinates width , (2 through height - 1) is left side
      if (side_exit1 != "right") {return false;}
    }
  } else {
    // exit1_gx is not on left or right edges , so exit1_gy should be on top or bottom edges
    if (exit1_gy == 1) {
      if (side_exit1 != "top") {return false;}
    } else if (exit1_gy == height) {
      if (side_exit1 != "bottom") {return false;}
    } else {
      /// exit1_gx,exit1_gy is not on any edge of the maze
      return false;
    }
  }
  
  /// check to make sure that the "side" of the the first grid coordinate's exit is on the outside edge
  if (exit2_gx == 1) {
    if (exit2_gy == 1) {
      // grid coordinate 1,1 is top left corner
      if ( (side_exit2 != "top") && (side_exit2 != "left") ) {return false;}
    } else if (exit2_gy == height) {
      // grid coordinate 1,height is bottom left corner
      if ( (side_exit2 != "bottom") && (side_exit2 != "left") ) {return false;}
    } else {
      // grid coordinates 1 , (2 through height - 1) is left side
      if (side_exit2 != "left") {return false;}
    }
  } else if (exit2_gx == width) {
    if (exit2_gy == 1) {
      // grid coordinate width,1 is top right corner
      if ( (side_exit2 != "top") && (side_exit2 != "right") ) {return false;}
    } else if (exit2_gy == height) {
      // grid coordinate width,height is bottom right corner
      if ( (side_exit2 != "bottom") && (side_exit2 != "right") ) {return false;}
    } else {
      // grid coordinates width , (2 through height - 1) is left side
      if (side_exit2 != "right") {return false;}
    }
  } else {
    // exit2_gx is not on left or right edges , so exit2_gy should be on top or bottom edges
    if (exit2_gy == 1) {
      if (side_exit2 != "top") {return false;}
    } else if (exit2_gy == height) {
      if (side_exit2 != "bottom") {return false;}
    } else {
      /// exit2_gx,exit2_gy is not on any edge of the maze
      return false;
    }
  }
  
  /// Set all outside edges of the maze to 'on = true' and 'checked = true'
  
  Edge* KE = NULL;
  
  for (int x = 1 ; x <= width ; x++) {
    // set top and bottom edges to on and checked
    if (this->GridCoordToEdgePtr(x , 1 , "top" , &KE)) {
      KE->SetEdgeStatus(true);
    } else {
      return false;// failed to get a pointer to the proper edge
    }
    if (this->GridCoordToEdgePtr(x , height , "bottom" , &KE)) {
      KE->SetEdgeStatus(true);
    } else {
      return false;// failed to get a pointer to the proper edge
    }
  }
  for (int y = 1 ; y <= height ; y++) {
    // set left and right edges to on and checked
    if (GridCoordToEdgePtr(1 , y , "left" , &KE)) {
      KE->SetEdgeStatus(true);
    } else {
      return false;// failed to get a pointer to the proper edge
    }
    if (GridCoordToEdgePtr(width , y , "right" , &KE)) {
      KE->SetEdgeStatus(true);
    } else {
      return false;// failed to get a pointer to the proper edge
    }
  }
  
  /// Now set the two exits to on = false , and checked = true
  if (GridCoordToEdgePtr(exit1_gx , exit1_gy , side_exit1 , &KE)) {
    KE->SetEdgeStatus(false);
  } else {
    return false;// failed to get a pointer to the proper edge
  }
  if (GridCoordToEdgePtr(exit2_gx , exit2_gy , side_exit2 , &KE)) {
    KE->SetEdgeStatus(false);
  } else {
    return false;// failed to get a pointer to the proper edge
  }
  
  
  return true;
}

bool KMaze::BuildShuffledWeightMap(const int& num_shuffles , ofstream& log_file) {
  
  /// Make a randomized array of pointers to the Edge array to store in the weight edge map
  // template <class Any>
  // Any* MakeRandomizedArrayCopy (Any* const ArrayToShuffle , const int& num_elements , const int& num_shuffles , ofstream& log_file) {
  Edge** EdgesPtrArray = new Edge*[total_edge_count];
  if (EdgesPtrArray == NULL) {return false;}
  
  int x = 0;
  for (x = 0 ; x < total_edge_count ; x++) {
    EdgesPtrArray[x] = &(AllEdgeArray[x]);
  }
  Edge** ShuffledEdgesPtrArray = MakeRandomizedArrayCopy(EdgesPtrArray , total_edge_count , num_shuffles , log_file);
  if (ShuffledEdgesPtrArray == NULL) {
    delete [] EdgesPtrArray;
    return false;
  }
  
  Edge* KE = NULL;
  
  wemm.clear();
  for (x = 0 ; x < total_edge_count ; x++) {
//    KE = &(AllEdgeArray[x]);
    KE = ShuffledEdgesPtrArray[x];
    wemm.insert(pair<int , Edge*>(KE->KWeight , KE));
  }
  
  
  delete [] EdgesPtrArray;
  delete [] ShuffledEdgesPtrArray;
  
  return true;
}



bool KMaze::IsLoopFormed (Vertice* VStart , Edge* EStart , bool start , bool* fail_ptr , BITMAP *bmp , bool* display , double& percent_complete) {
  /// This function lets you know whether adding the edge EStart to the v/e/v/... path links back to itself
  ///   through any connecting edges / vertices / edges if the edges data member 'on' is true
  static vector<Vertice*> VertsOnPath;
  static vector<Edge*> EdgesOnPath;
  static bool pause = false;
  
  static stringstream progress;
  
  int v_col = makecol(255,0,255);
  int e_on_col = makecol(255,255,0);
  int e_off_col = makecol(255,0,0);
  if (start) {
    if (*display) {
      clear_to_color(bmp , makecol(0,0,0));
      DrawAllEdges(bmp , makecol(0,0,255) , makecol(0,0,0) , false);
      textout_ex(bmp , font , "P key to un/pause" , 10 , 4 , makecol(255,255,255) , -1);
      textout_ex(bmp , font , "When paused , press any key other than P or ESC to see the next step" , 10 , 16 , makecol(255,255,255) , -1);
      textout_ex(bmp , font , "ESC key to skip displaying path calculation" , 10 , 28 , makecol(255,255,255) , -1);
      textout_ex(bmp , font , "Blue Edges are already turned on" , 10 , 40 , makecol(255,255,255) , -1);
      textout_ex(bmp , font , "Red Edge is being tested to see if turning it on forms a loop" , 10 , 52 , makecol(255,255,255) , -1);
    }
    progress.str("");
    progress << "Percent Complete = " << percent_complete << "        ";
    textout_ex(bmp , font , (progress.str()).c_str() , (bmp->w / 2) - 104 , bmp->h - 20 , makecol(255,255,255) , makecol(0,0,0));
    blit(bmp , screen , 0 , 0 , 0 , 0 , bmp->w , bmp->h);
//    pause = false;
    VertsOnPath.clear();
    EdgesOnPath.clear();
    *fail_ptr = false;
  }
  
  Vertice* Vnext = NULL;
  Vertice* V = NULL;
  
  if (EStart->vpt1 == VStart) {
    Vnext = EStart->vpt2;
  } else if (EStart->vpt2 == VStart) {
    Vnext = EStart->vpt1;
  } else {
    *fail_ptr = true;
    return false;
  }
  
  VertsOnPath.push_back(VStart);
  EdgesOnPath.push_back(EStart);
  
  int user_key = 0;
  
  if (*display) {
    for (unsigned int i = 0 ; i < VertsOnPath.size() ; i++) {
//      (VertsOnPath[i])->Draw(bmp , makecol(255,0,255) , 12);
      V = VertsOnPath[i];
      V->Draw(bmp , v_col , 3);
    }
    for (unsigned int i = 0 ; i < EdgesOnPath.size() ; i++) {
//      (EdgesOnPath[i])->Draw(bmp , makecol(255,255,0) , makecol(255,0,0) , true);
      (EdgesOnPath[i])->Draw(bmp , e_on_col , e_off_col , true);
    }
    blit(bmp , screen , 0 , 0 , 0 , 0 , bmp->w , bmp->h);
    if (key[KEY_ESC]) {
      textout_centre_ex(bmp , font , "Finishing Maze Calculation" , bmp->w / 2 , bmp->h - 30 , makecol(255,255,255) , -1);
      blit(bmp , screen , 0 , 0 , 0 , 0 , bmp->w , bmp->h);
      *display = false;
      rest(200);
    }
    if (key[KEY_P]) {
      pause = !pause;
      rest(200);
    }
    if (pause) {
      clear_keybuf();
      user_key = readkey() >> 8;
      clear_keybuf();
      if (user_key == KEY_ESC) {
        textout_centre_ex(bmp , font , "Finishing Maze Calculation" , bmp->w / 2 , bmp->h - 36 , makecol(255,255,255) , -1);
        blit(bmp , screen , 0 , 0 , 0 , 0 , bmp->w , bmp->h);
        *display = false;
        rest(200);
      }
      if (user_key == KEY_P) {
        pause = !pause;
        rest(200);
      }
    } else {
//      rest(10);
      rest(0);
    }
  }
  
  unsigned int v_edge_num = 0;
  unsigned int path_v_edge_num = 0;
  
  /// check to make sure the next Vertice is not already on the path , if it is then a loop has been formed
  for (path_v_edge_num = 0 ; path_v_edge_num < VertsOnPath.size() ; path_v_edge_num++) {
    if (Vnext == VertsOnPath[path_v_edge_num]) {
      return true;
    }
  }
  /// for each edge on the next vertice that is on , recall this function
  for (v_edge_num = 0 ; v_edge_num < Vnext->kepvec.size() ; v_edge_num++) {
    if (Vnext->kepvec[v_edge_num]->on == true) {
      if (Vnext->kepvec[v_edge_num] != EStart) {
        if (IsLoopFormed(Vnext , Vnext->kepvec[v_edge_num] , false , fail_ptr , bmp , display , percent_complete)) {return true;} 
      }
    }
  }
  return false;
}


bool KMaze::KruskalizeGridEdges (ofstream& log_file , BITMAP* bmp , bool display) {
  // need to know how many unique weights have been assigned to the entire group of edges
  // # actually , only need to make a sorted array of pointers
  
  /// Build shuffled weight edge multi map
  const int times_to_shuffle = 3;
  if (!(BuildShuffledWeightMap(times_to_shuffle , log_file))) {return false;}
  
  Edge* KE = NULL;
  
  bool loop_formed_vpt1 = false;
  bool loop_formed_vpt2 = false;
  bool loop_formed = false;
  bool loop_check_failed = false;
  Vertice* V = NULL;
  
  typedef multimap<int , Edge* , ltint>::iterator wemm_it;
  wemm_it wm_it;
  
  int edges_checked = 0;
  double percent_complete = 0.0;
  
  log_file << endl;
  log_file << "Kruskalizing grid edges , there are " << total_edge_count << " Edges to check" << endl;
  log_file << "-------------------------------------------------------------" << endl;
  for(wm_it = wemm.begin() ; wm_it != wemm.end() ; wm_it++) {
    KE = wm_it->second;
    
    if (!(KE->checked)) {
      loop_formed = false;
      log_file << "Checking Edge " << KE;
      V = KE->vpt1;
      percent_complete = static_cast<double>(100*edges_checked) / static_cast<double>(total_edge_count);
      loop_formed_vpt1 = IsLoopFormed(V , KE , true , &loop_check_failed , bmp , &display , percent_complete);
      if (!loop_formed_vpt1) {if (loop_check_failed) {return false;}}
      
//      V = KE->vpt2;
//      loop_formed_vpt2 = IsLoopFormed(V , KE , true , &loop_check_failed , bmp , &display);
//      if (!loop_formed_vpt2) {if (loop_check_failed) {return false;}}
      
      if ((loop_formed_vpt1) || (loop_formed_vpt2)) {loop_formed = true;}
      
      /// If there is a loop formed by turning on KE then KE can not be turned on
      /// If there is no loop formed by turning on KE then KE can be turned on
      KE->SetEdgeStatus(!loop_formed);
      log_file << " , Setting Edge " << KE << " to ";
      if (loop_formed) {log_file << "off";}
      if (!loop_formed) {log_file << "on";}
      log_file << " , loop formed by edges connected to vpt1 = "
                << loop_formed_vpt1 << endl;
                // << " , loop formed by edges connected to vpt2 = " << loop_formed_vpt2 << endl;
    } else {
      log_file << "This Edge " << KE << " was already checked before the Kruskal Algorithm was applied" << endl;
    }
    edges_checked++;
  }
  
  return true;
}

bool KMaze::KruskalizeGridEdgesByVTrees(ofstream& log_file , BITMAP* bmp , bool display) {
  
  
  /// Build shuffled weight edge multi map
  const int times_to_shuffle = 3;
  if (!(BuildShuffledWeightMap(times_to_shuffle , log_file))) {return false;}
  
  
  vector< vector<Vertice*> > VTrees;
  VTrees.clear();
  int Num_VTrees = 0;
  vector< vector<Vertice*> >::iterator VTrees_iterator;
  vector<Vertice*>* VTree_ptr = NULL;
  vector<Vertice*>* VTree_join_to_ptr = NULL;
  vector<Vertice*>* VTree_join_from_ptr = NULL;
  int VTree_size = 0;
  
  vector<Vertice*> temp_VTree;
  temp_VTree.clear();
  
  Edge* KE = NULL;
  Vertice* V1 = NULL;
  Vertice* V2 = NULL;
  bool v1ot = false;
  bool v2ot = false;
  int v1_tree_num = 0;
  int v2_tree_num = 0;
  
  
  /// Separate the Weight Edge Multi Map into two vectors of Edge pointers , one checked and one unchecked
  std::vector<Edge*> checked_edges;
  std::vector<Edge*> unchecked_edges;
  checked_edges.clear();
  unchecked_edges.clear();
  
//    multimap<int , Edge* , ltint> wemm;// Weight Edge Multi Map
  multimap<int , Edge* , ltint>::iterator wmit;
  
  for (wmit = wemm.begin() ; wmit != wemm.end() ; wmit++) {
    KE = wmit->second;
    if (KE->checked) {
      checked_edges.push_back(KE);
    } else {
      unchecked_edges.push_back(KE);
    }
  }
  
  int num_edges_checked = 0;
  double total_num_edges = static_cast<double>(total_edge_count);
  int percent_complete = 0;
  int previous_percent = 0;
  
  char outputstring[100];
  char numberstring[50];
  for (int x = 0 ; x < 100 ; x++) {outputstring[x] = '\0';}
  for (int x = 0 ; x < 50 ; x++) {numberstring[x] = '\0';}
  const int white = makecol(255,255,255);
  const int black = makecol(0,0,0);
  
  /// first find existing trees of edges(actually vertices) that are turned on and checked
  ///   of those that are on the AllEdgeArray
    // for all edges that are already checked , 
    //     - if neither is part of a tree , make a new tree and add them both to it
    //     - if either is on a tree , but not both , add the vertice that is not on a tree to the tree the vertice that is on is on
    //     - if both are on a tree
    //       - if they are on the same tree , the edge loops back to it's own tree ,
    //           which is not allowed unless the edge has already been checked (user set)
    //       - if they are on different trees , then link the two trees together since they are one
  for (unsigned int ce = 0 ; ce < checked_edges.size() ; ce++) {
    KE = checked_edges[ce];
    if (!(KE->on)) {continue;}// do not add edges that are off to the tree of edges that are on
    V1 = KE->vpt1;
    V2 = KE->vpt2;
    v1ot = false;
    v2ot = false;
    Num_VTrees = VTrees.size();
    for (int avt = 0 ; avt < Num_VTrees ; avt++) {
      VTree_ptr = &(VTrees[avt]);
      VTree_size = VTree_ptr->size();
      for (int svt = 0 ; svt < VTree_size ; svt++) {
        if (!v1ot) {// if vertice 1 has not been found on any tree , compare to current node on current tree
          if (V1 == (*VTree_ptr)[svt]) {
            v1ot = true;
            v1_tree_num = avt;
          }
        }
        if (!v2ot) {// if vertice 2 has not been found on any tree , compare to current node on current tree
          if (V2 == (*VTree_ptr)[svt]) {
            v2ot = true;
            v2_tree_num = avt;
          }
        }
        if (v1ot && v2ot) {
          // break out of tree and node search loop since both vertices have already been found on the trees
          svt = VTree_size;
          avt = Num_VTrees;
          // since this tree building method checks all the edges in order , each vertice can only be on one tree
          //   , rather , a vertice will only be added if it is not on a tree , so there can be no doubles
        }
      }
    }

    if ((!v1ot) && (!v2ot)) {
      // neither vertice is on any tree , so add both vertices of this edge to a new tree
      temp_VTree.push_back(V1);
      temp_VTree.push_back(V2);
      VTrees.push_back(temp_VTree);
      temp_VTree.clear();
    } else {// at least one vertice on a tree
      if (v1ot && v2ot) {// both on a tree
        if (v1_tree_num == v2_tree_num) {
          // both on same tree , can not add edge otherwise the tree links back to itself , so turn edge off and set to checked
          /// except that these are edges that have already been checked so add them to the tree anyway
          /// When the unchecked edges are looked over , it will still work properly
          VTrees[v1_tree_num].push_back(V1);
          VTrees[v2_tree_num].push_back(V2);
        } else {
          // both vertices on different trees , so link the trees together
          // push back all the Vertice pointers from one tree onto the other and delete the unneeded tree
          VTree_join_to_ptr = &(VTrees[v1_tree_num]);
          VTree_join_from_ptr = &(VTrees[v2_tree_num]);
          for (unsigned int j = 0 ; j < VTrees[v2_tree_num].size() ; j++) {
//              VTrees[v1_tree_num].push_back((VTrees[v2_tree_num])[j])
            VTree_join_to_ptr->push_back((*VTree_join_from_ptr)[j]);
          }
          VTree_join_from_ptr->clear();
          // need an iterator to VTrees[v2_tree_num] so it can be erased
          VTrees_iterator = VTrees.begin();
          VTrees_iterator += v2_tree_num;
          VTrees.erase(VTrees_iterator);
        }
      } else {
        // only one vertice is on a tree , so add other vertice to the tree the first one is on
        if (v1ot) {
          // vertice one is on a tree and vertice two is not , so add vertice two to the tree vertice one is on
          VTrees[v1_tree_num].push_back(V2);
        }
        if (v2ot) {
          // vertice two is on a tree and vertice one is not , so add vertice one to the tree vertice two is on
          VTrees[v2_tree_num].push_back(V1);
        }
        
      }
      
    }
//    log_file << num_edges_checked++ << " edges checked , Edge " << KE << " was already checked and was " << KE->on << endl;
    if (display) {
      percent_complete = static_cast<int>(static_cast<double>(100*num_edges_checked) / total_num_edges);
      if (percent_complete > previous_percent) {
        previous_percent = percent_complete;
        strcpy(outputstring , "Percent Complete ");
        itoa(percent_complete , numberstring , 10);
        strcat(outputstring , numberstring);
        strcat(outputstring , "  ");
        textout_ex(bmp , font , outputstring , 320 , 566 , white , black);
        blit(bmp , screen , 0 , 0 , 0 , 0 , bmp->w , bmp->h);
      }
    }
  } /// All checked edges have been made into trees
  
//  log_file << endl;
//  log_file << endl << "Checking unchecked edges to see if they should be turned on" << endl;
  
  // second go through the Weight Edge Multi Map and turn on edges that do not loop back (do not connect back to same tree)
  /// The unchecked edges vector has been made from the Weight Map and retains the weight order from lowest to highest
  ///   so go through the vector instead
  for (unsigned int uce = 0 ; uce < unchecked_edges.size() ; uce++) {
    KE = unchecked_edges[uce];
    // need to check all unchecked edges to find whether to set them on or off
    V1 = KE->vpt1;
    V2 = KE->vpt2;
    v1ot = false;
    v2ot = false;
    Num_VTrees = VTrees.size();
    for (int avt = 0 ; avt < Num_VTrees ; avt++) {
      VTree_ptr = &(VTrees[avt]);
      VTree_size = VTree_ptr->size();
      for (int svt = 0 ; svt < VTree_size ; svt++) {
        if (!v1ot) {// if vertice 1 has not been found on any tree , compare to current node on current tree
          if (V1 == (*VTree_ptr)[svt]) {
            v1ot = true;
            v1_tree_num = avt;
          }
        }
        if (!v2ot) {// if vertice 2 has not been found on any tree , compare to current node on current tree
          if (V2 == (*VTree_ptr)[svt]) {
            v2ot = true;
            v2_tree_num = avt;
          }
        }
        if (v1ot && v2ot) {
          // break out of tree and node search loop since both vertices have already been found on the trees
          svt = VTree_size;
          avt = Num_VTrees;
          // since this tree building method checks all the edges in order , each vertice can only be on one tree
          //   , rather , a vertice will only be added if it is not on a tree , so there can be no doubles
        }
      }
    }

    /// Turn on edges that do not loop back , and add vertices to appropriate trees / combine trees
    if ((!v1ot) && (!v2ot)) {
      // neither vertice is on any tree , so add both vertices of this edge to a new tree and turn the edge on
      temp_VTree.push_back(V1);
      temp_VTree.push_back(V2);
      VTrees.push_back(temp_VTree);
      temp_VTree.clear();
      KE->SetEdgeStatus(true);
    } else {// at least one vertice on a tree
      if (v1ot && v2ot) {// both on a tree
        if (v1_tree_num == v2_tree_num) {
          // both vertices already on same tree , can not add edge otherwise the tree links back to itself ,
          //    so turn edge off and set to checked. Make sure not to add vertices to any tree
          KE->SetEdgeStatus(false);
///          VTrees[v1_tree_num].push_back(V1);
///          VTrees[v2_tree_num].push_back(V2);
        } else {
          // both vertices on different trees , so link the trees together and turn the edge on
          // push back all the Vertice pointers from one tree onto the other and delete the unneeded tree
          KE->SetEdgeStatus(true);
          VTree_join_to_ptr = &(VTrees[v1_tree_num]);
          VTree_join_from_ptr = &(VTrees[v2_tree_num]);
          for (unsigned int j = 0 ; j < VTrees[v2_tree_num].size() ; j++) {
//              VTrees[v1_tree_num].push_back((VTrees[v2_tree_num])[j])
            VTree_join_to_ptr->push_back((*VTree_join_from_ptr)[j]);
          }
          VTree_join_from_ptr->clear();
          // need an iterator to VTrees[v2_tree_num] so it can be erased
          VTrees_iterator = VTrees.begin();
          VTrees_iterator += v2_tree_num;
          VTrees.erase(VTrees_iterator);
        }
      } else {
        // at most only one vertice is on a tree , so add other vertice to the tree the first one is on and turn the edge on
        if (v1ot) {
          // vertice one is on a tree and vertice two is not , so add vertice two to the tree vertice one is on
          VTrees[v1_tree_num].push_back(V2);
        }
        if (v2ot) {
          // vertice two is on a tree and vertice one is not , so add vertice one to the tree vertice two is on
          VTrees[v2_tree_num].push_back(V1);
        }
        KE->SetEdgeStatus(true);
      }
      
    }
//    log_file << num_edges_checked++ << " edges checked , Edge " << KE << " has been turned " << KE->on << endl;
    if (display) {
      percent_complete = static_cast<int>(static_cast<double>(100*num_edges_checked) / total_num_edges);
      if (percent_complete > previous_percent) {
        previous_percent = percent_complete;
        strcpy(outputstring , "Percent Complete ");
        itoa(percent_complete , numberstring , 10);
        strcat(outputstring , numberstring);
        strcat(outputstring , "  ");
        textout_ex(bmp , font , outputstring , 320 , 566 , white , black);
        blit(bmp , screen , 0 , 0 , 0 , 0 , bmp->w , bmp->h);
      }
    }
  } /// All unchecked edges have been made into trees and turned on or off appropriately
  return true;
}


bool KMaze::GridCoordToVerticePtr(int gx , int gy , string side , Vertice** vptr_address) {
  /// returns true if successful in setting *vptr_address to the appropriate edge pointer
  /// One based (3 wide = 1,2,3 and 4 tall = 1,2,3,4)
  if (vptr_address == NULL) {return false;}
  if ((gx < 1) || (gx > width)) {return false;}
  if ((gy < 1) || (gy > height)) {return false;}
  
  if ( !(((side == "topleft") || (side == "topright")) || ((side == "bottomleft") || (side == "bottomright"))) ) {return false;}
  
  unsigned int vertice_element = 0;
  
  // find "topleft" element first , adjust accordingly
  vertice_element = (gy - 1)*(width + 1) + (gx - 1);
  if (side == "topright") {
    vertice_element += 1;
  }
  if (side == "bottomleft") {
    vertice_element += (width + 1);
  }
  if (side == "bottomright") {
    vertice_element += (width + 1) + 1;
  }
  *vptr_address = &(VerticeArray[vertice_element]);
  
  return true;
}

bool KMaze::GridCoordToEdgePtr(int gx , int gy , string side , Edge** eptr_address) {
  /// returns true if successful in setting *eptr_address to the appropriate edge pointer
  /// One based (3 wide = 1,2,3 and 4 tall = 1,2,3,4)
  if (eptr_address == NULL) {return false;}
  if ((gx < 1) || (gx > width)) {return false;}
  if ((gy < 1) || (gy > height)) {return false;}
  
  bool left_right = false;
  bool top_bottom = false;
  bool right = false;
  bool bottom = false;
  
  if ((side == "left") || (side == "right")) {
    left_right = true;
    if (side == "right") {right = true;}
  }
  if ((side == "top") || (side == "bottom")) {
    top_bottom = true;
    if (side == "bottom") {bottom = true;}
  }
  
  if (!(left_right) && !(top_bottom)) {return false;}
  
  unsigned int edge_element = 0;
  
  if (top_bottom) {
    edge_element = (gy - 1)*(2*width + 1) + (gx - 1);
    if (bottom) {edge_element += (2*width + 1);}
    /// test case (3 X 2 grid , grid coordinate 1,1)
    ///   top edge element = 0*(2*3 + 1) + (1 - 1) = 0
    ///   bottom element   = (0) + (2*3 + 1) = 7
    /// test case (3 X 2 grid , grid coordinate 3,2)
    ///   top edge element = (2-1)*(2*3 + 1) + (3-1) = 7 + 2 = 9
    ///   bottom element   = (9) + (2*3 + 1) = 16
    /// Max number of elements = 2*(w*h) + w + h = 2*(3*2) + 3 + 2 = 17 so 16 is correct
  }
  if (left_right) {
    edge_element = width + (gy - 1)*(2*width + 1) + (gx - 1);
    if (right) {edge_element += 1;}
    /// test case (3 X 2 grid , grid coordinate 2,1)
    ///   left edge element  = 3 + (1-1)*(2*3 + 1) + (2-1) = 4
    ///   right edge element = (4) + 1 = 5
  }
  
  *eptr_address = &(AllEdgeArray[edge_element]);
  return true;
}

void KMaze::DrawVertice(int v_num , BITMAP* bmp , int v_color , int v_width) {
  if ((v_num < 0) || (v_num >= total_vertice_count)) {return;}
  if (bmp == NULL) {return;}
  
  Vertice* const V = &(VerticeArray[v_num]);
  
  V->Draw(bmp , v_color , v_width);
}

void KMaze::IsoMetricDrawVertice(const double& z_angle_rad , int v_num , BITMAP* bmp , int v_color , int v_width) {
  if ((v_num < 0) || (v_num >= total_vertice_count)) {return;}
  if (bmp == NULL) {return;}
  
  Vertice* const V = &(VerticeArray[v_num]);
  
  V->IsoMetricDraw(z_angle_rad , bmp , v_color , v_width);
}



void KMaze::DrawVerticeEdges(int v_num , BITMAP* bmp , int on_color , int off_color , bool draw_off_edges) {
  if ((v_num < 0) || (v_num >= total_vertice_count)) {return;}

  Edge* KE = NULL;

  Vertice* Vp = NULL;
  Vp = &(VerticeArray[v_num]);

  for (unsigned int ve = 0 ; ve < Vp->kepvec.size() ; ve++) {
    KE = Vp->kepvec[ve];
    KE->Draw(bmp , on_color , off_color , draw_off_edges);
  }
}



void KMaze::IsoMetricDrawVerticeEdges(const double& z_angle_rad , int v_num , BITMAP* bmp , int on_color , int off_color , bool draw_off_edges) {
  if ((v_num < 0) || (v_num >= total_vertice_count)) {return;}
  
  Edge* KE = NULL;
  Vertice* Vp = NULL;
  Vp = &(VerticeArray[v_num]);
  
  for (unsigned int ve = 0 ; ve < Vp->kepvec.size() ; ve++) {
    KE = Vp->kepvec[ve];
    KE->IsoMetricDraw(z_angle_rad , bmp , on_color , off_color , draw_off_edges);
  }
}


void KMaze::DrawAllVertices(BITMAP* bmp , int v_color , int v_width) {
  Vertice* V = NULL;
  for (int i = 0 ; i < total_vertice_count ; i++) {
    V = &(VerticeArray[i]);
    V->Draw(bmp , v_color , v_width);
  }
}

void KMaze::IsoMetricDrawAllVertices(const double& z_angle_rad , BITMAP* bmp , int v_color , int v_width) {
  Vertice* V = NULL;
  for (int i = 0 ; i < total_vertice_count ; i++) {
    V = &(VerticeArray[i]);
    V->IsoMetricDraw(z_angle_rad , bmp , v_color , v_width);
  }
}

void KMaze::DrawEdge(const int& edge_num , BITMAP* bmp , const int& on_color , const int& off_color , const bool& draw_off_edges) {
  if ((edge_num < 0) || (edge_num >= total_edge_count)) {return;}
  Edge* KE = NULL;
  KE = &(AllEdgeArray[edge_num]);
  KE->Draw(bmp , on_color , off_color , draw_off_edges);
}

void KMaze::IsoMetricDrawEdge(const double& z_angle_rad , const int& edge_num , BITMAP* bmp , const int& on_color , const int& off_color , const bool& draw_off_edges) {
  if ((edge_num < 0) || (edge_num >= total_edge_count)) {return;}
  Edge* KE = NULL;
  KE = &(AllEdgeArray[edge_num]);
  KE->IsoMetricDraw(z_angle_rad , bmp , on_color , off_color , draw_off_edges);
}

void KMaze::DrawAllEdges(BITMAP* bmp , int on_color , int off_color , bool draw_off_edges) {
  Edge* KE = NULL;
  for (int edge_count = 0 ; edge_count < total_edge_count ; edge_count++) {
    KE = &(AllEdgeArray[edge_count]);
    KE->Draw(bmp , on_color , off_color , draw_off_edges);
  }
}
void KMaze::IsoMetricDrawAllEdges(const double& z_angle_rad , BITMAP* bmp , int on_color , int off_color , bool draw_off_edges) {
  Edge* KE = NULL;
  for (int edge_count = 0 ; edge_count < total_edge_count ; edge_count++) {
    KE = &(AllEdgeArray[edge_count]);
    KE->IsoMetricDraw(z_angle_rad , bmp , on_color , off_color , draw_off_edges);
  }
}

/// End KMaze class member functions
