
#ifndef KMaze_H
#define KMaze_H

/// int rand0toN(const int& N);
/// int rand1toN(const int& N);

/// template <class Any>
/// Any* MakeRandomizedArrayCopy (Any* const ArrayToShuffle , const int& num_elements , const int& num_shuffles , ofstream& log_file) {

/// class Vertice
/// class Edge
/// class KMaze


#include <vector>
#include <map>
#include <string>

#include <iostream>
#include <fstream>

#include <allegro.h>

#include "Vertices.hpp"
#include "Edges.hpp"

using std::vector;
using std::multimap;
using std::pair;
using std::string;

using std::endl;
using std::ofstream;


/// Less than integer bool function for use in the KMaze weight edge* multimap
struct ltint
{
  bool operator()(int i1 , int i2) const
  {
    return (i1 < i2);
  }
};

/// ########################## KMaze class ####################################
class KMaze {
  public :
    /// data
    int width;// number of grid squares wide
    int height;// number of grid squares high
    int sg_width;// width of single grid column
    int sg_height;// height of single grid row
    int top_left_x;// top left corner x position , top right is tlx + width*sg_width - 1
    int top_left_y;// top left corner y position , bottom left is tly + height*sg_height - 1
    
    int total_vertice_count;
    Vertice* VerticeArray;
    
    int total_edge_count;
    Edge* AllEdgeArray;
    
    multimap<int , Edge* , ltint> wemm;// Weight Edge Multi Map
    Edge** EdgesOnArray;// for optimizations?
    
    /// methods
    KMaze();
    ~KMaze();
    
    bool InitializeGrid(int w , int h , int tlx , int tly , int grid_z_level , int xwid , int yht);
    void DestroyGrid();
    
    void ResetGridEdges ();
    bool SetMazeBorderOn ();
    bool SetMazeBorderOn (int exit1_gx , int exit1_gy , string side_exit1 , int exit2_gx , int exit2_gy , string side_exit2);
    bool SetExit(int exit_gx , int exit_gy , string side_exit);
    
    bool BuildShuffledWeightMap(const int& num_shuffles , ofstream& log_file);
    bool IsLoopFormed (Vertice* VStart , Edge* EStart , bool start , bool* fail_ptr , BITMAP *bmp , bool* display , double& percent_complete);
    bool KruskalizeGridEdges (ofstream& log_file , BITMAP* bmp , bool display);
    
    bool KruskalizeGridEdgesByVTrees(ofstream& log_file , BITMAP* bmp , bool display);
    
    bool GridCoordToVerticePtr(int gx , int gy , string side , Vertice** vptr_address);
    bool GridCoordToEdgePtr(int gx , int gy , string side , Edge** eptr_address);
//    void XYfromGrid(const int& grx , const int& gry , int* const rx , int* const ry);
    void KMaze::XYfromGrid(const int& grx , const int& gry , int* const rx , int* const ry) {
      *rx = top_left_x + (grx-1)*sg_width;
      *ry = top_left_y + (gry-1)*sg_height;
    }
    
    void DrawVertice(int v_num , BITMAP* bmp , int v_color , int v_width);
    void IsoMetricDrawVertice(const double& z_angle_rad , int v_num , BITMAP* bmp , int v_color , int v_width);
    
    void DrawVerticeEdges(int v_num , BITMAP* bmp , int on_color , int off_color , bool draw_off_edges);
    void IsoMetricDrawVerticeEdges(const double& z_angle_rad , int v_num , BITMAP* bmp , int on_color , int off_color , bool draw_off_edges);
    
    void DrawAllVertices(BITMAP* bmp , int v_color , int v_width);
    void IsoMetricDrawAllVertices(const double& z_angle_rad , BITMAP* bmp , int v_color , int v_width);
    
    void DrawEdge(const int& edge_num , BITMAP* bmp , const int& on_color , const int& off_color , const bool& draw_off_edges);
    void IsoMetricDrawEdge(const double& z_angle_rad , const int& edge_num , BITMAP* bmp , const int& on_color , const int& off_color , const bool& draw_off_edges);
    
    void DrawAllEdges(BITMAP* bmp , int on_color , int off_color , bool draw_off_edges);
    void IsoMetricDrawAllEdges(const double& z_angle_rad , BITMAP* bmp , int on_color , int off_color , bool draw_off_edges);
    
};


#endif




