
#ifndef Shuffle_H
#define Shuffle_H

#include <fstream>
#include <iostream>

using std::endl;
using std::ofstream;

#include "randomizer.hpp"

// #################### Random number functions ########################

/// Template function to make a new array that is a shuffled copy of the array 
///   that is passed to it.
/// Ideally , should be used with an array of pointers but will work with an array of objects too
///
/// Don't forget to delete [] the array when you're done with it

template <class Any>
Any* MakeRandomizedArrayCopy (Any* const ArrayToShuffle , const int& num_elements , const int& num_shuffles , ofstream& log_file) {
  
  if (ArrayToShuffle == NULL) {return NULL;}
  if (num_elements < 2) {return NULL;}
  if (num_shuffles < 1) {return NULL;}
  
  Any* RandomArray = new Any[num_elements];
  if (RandomArray == NULL) {return NULL;}
  
  srand(time(NULL));
  
  std::vector<int> unpicked_elements;
  unpicked_elements.clear();
  unpicked_elements.resize(num_elements);
  
  std::vector<int> picked_elements;
  picked_elements.clear();
  picked_elements.resize(num_elements);
  
  std::vector<int>::iterator upv_it;
  
  for (int i = 0 ; i < num_elements  ; i++) {
    unpicked_elements[i] = i;
  }
  
  int elements_left = 0;
  int random_element = 0;
  
  /// Shuffle the order that the elements will be placed into the randomized array
  for (int s = 1 ; s <= num_shuffles ; s++) {
    elements_left = num_elements;
    for (int i = 0 ; i < num_elements ; i++) {
      upv_it = unpicked_elements.begin();
      random_element = random_0toN(elements_left - 1);
      upv_it += random_element;
      
      picked_elements[i] = unpicked_elements[random_element];
      unpicked_elements.erase(upv_it);
      
      elements_left--;
    }
    unpicked_elements = picked_elements;// to repeat the shuffle
  }
  
  /// Build the shuffled copy of the array
//  log_file << "The shuffled order of the array is : " << endl;
  for (int i = 0 ; i < num_elements ; i++) {
//    log_file << picked_elements[i] << " ";
    RandomArray[i] = ArrayToShuffle[picked_elements[i]];
  }
//  log_file << endl;
  
  return RandomArray;
}

#endif
