

#include <cmath>

#include "Vertices.hpp"


/// Vertice class memeber functions
Vertice::Vertice() {
  x = 0;
  y = 0;
  kepvec.clear();
}

Vertice::~Vertice() {
  if (kepvec.size() > 0) {
    kepvec.clear();
  }
}

void Vertice::assign_edge(Edge* const ke) {
  if (ke == NULL) {return;}
  
  bool unique = true;
  for (unsigned int i = 0 ; i < kepvec.size() ; i++) {
    if (ke == kepvec[i]) {
      unique = false;
    }
  }
  if (unique) {
    kepvec.push_back(ke);
  }
}

void Vertice::clear_edges() {
  if (kepvec.size() > 0) {
    kepvec.clear();
  }
}

bool Vertice::check_for_edge(Edge* const ke) {
  if (ke == NULL) {return false;}
  
  for (unsigned int i = 0 ; i < kepvec.size() ; i++) {
    if (ke == kepvec[i]) {return true;}
  }
  return false;
}

void Vertice::Draw(BITMAP* bmp , int v_color , int v_wid) {
  const int hwid = v_wid / 2;
  const int ohwid = v_wid - hwid;
  
  rectfill(bmp , x - hwid , y - hwid , x + ohwid , y + ohwid , v_color);
}
void Vertice::IsoMetricDraw(const double& z_angle_rad , BITMAP* bmp , int v_color , int v_wid) {
  const int hwid = v_wid / 2;
  const int ohwid = v_wid - hwid;
  
  const int x_offset = static_cast<int>(static_cast<double>(z)*cos(z_angle_rad));
  const int y_offset = static_cast<int>(static_cast<double>(z)*sin(z_angle_rad));
  
  rectfill(bmp , x + x_offset - hwid , y + y_offset - hwid , x + x_offset + ohwid , y + y_offset + ohwid , v_color);
  
}

void Vertice::IsoMetricDrawBar(const double& z_angle_rad , BITMAP* bmp , int v_color , int v_wid) {
  
  /// vci = Vertex Color Intensity
  int vci = getr(v_color) + getg(v_color) + getb(v_color);
  int vci_level = vci / 256;// should be 0 , 1 , or 2
  
  int shadow_color = 0;
  int normal_color = 0;
  int highlight_color = 0;
  int r,g,b;
  
  r = getr(v_color);
  g = getg(v_color);
  b = getb(v_color);
  
  switch(vci_level) {
    case 0 :
      shadow_color = v_color;
      normal_color = makecol(r + (255-r)/3 , g + (255-g)/3 , b + (255-b)/3);
      highlight_color = makecol(r + (2*(255-r))/3 , g + (2*(255-g))/3 , b + (2*(255-b))/3);
      break;
    case 1 :
      normal_color = v_color;
      shadow_color = makecol(r/2 , g/2 , b/2);
      highlight_color = makecol(r + (255-r)/2 , g + (255-g)/2 , b + (255-b)/2);
      break;
    case 2 :
      highlight_color = v_color;
      normal_color = makecol((2*r)/3 , (2*g)/3 , (2*b)/3);
      shadow_color = makecol((1*r)/3 , (1*g)/3 , (1*b)/3);
      break;
  }
  
  int tlx,tly,brx,bry;
  int x_offset , y_offset;
  int z_dist = z;
//  for (int z_dist = z ; z_dist > 0 ; z_dist--) {
  for (int i = abs(z) ; i > 0 ; i--) {
    x_offset = static_cast<int>(static_cast<double>(z_dist)*cos(z_angle_rad));
    y_offset = static_cast<int>(static_cast<double>(z_dist)*sin(z_angle_rad));
    tlx = x + x_offset - v_wid/2;
    tly = y + y_offset - v_wid/2;
    brx = tlx + v_wid - 1;
    bry = tly + v_wid - 1;
    triangle(bmp , brx , tly , tlx , bry , brx , bry , highlight_color);
    triangle(bmp , brx , tly , tlx , bry , tlx , tly , shadow_color);
    if (z > 0) {z_dist--;} else {z_dist++;}
  }
  Draw(bmp , normal_color , v_wid);
}

/// End Vertice class member functions
